"""Module containing a class for an in-memory token cache"""
from azol.caches.azol_cache import AzolCache

class InMemoryTokenCache( AzolCache ):
    """Class inheriting from AzolCache that provides an in-memory token cache"""
